# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 The NebulaGraph Authors. All rights reserved.
from __future__ import annotations

from ngdi.config import NebulaGraphConfig

# SPARK DEFAULTS
DEFAULT_SHUFFLE_PARTITIONS = 5
DEFAULT_EXECUTOR_MEMORY = "8g"
DEFAULT_DRIVER_MEMORY = "4g"

# NebulaGraph Algorithm Path
NEBULA_ALGO = "com.vesoft.nebula.algorithm"
NEBULA_ALGO_CONFIG = f"{NEBULA_ALGO}.config"
NEBULA_ALGO_LIB = f"{NEBULA_ALGO}.lib"

# NebulaGraph Spark Connector Path
NEBULA_SPARK_CONNECTOR = "com.vesoft.nebula.connector"
NEBULA_SPARK_CONNECTOR_DATASOURCE = f"{NEBULA_SPARK_CONNECTOR}.NebulaDataSource"

# NebulaGraph Algorithm Default Config
ENCODE_VERTEX_ID = True


class BaseEngine(object):
    def __init__(self, config):
        pass

    def __str__(self):
        return f"BaseEngine: {self.type}"

    def prepare():
        raise NotImplementedError


class SparkEngine(BaseEngine):
    def __init__(self, config):
        self.type = "spark"
        self.config = config
        self.shuffle_partitions = DEFAULT_SHUFFLE_PARTITIONS
        self.executor_memory = DEFAULT_EXECUTOR_MEMORY
        self.driver_memory = DEFAULT_DRIVER_MEMORY
        self.encode_vertex_id = ENCODE_VERTEX_ID
        self.parse_config()

        from pyspark.sql import SparkSession

        self.spark = (
            SparkSession.builder.appName("NebulaGraph Data Intelligence")
            .config("spark.sql.shuffle.partitions", self.shuffle_partitions)
            .config("spark.executor.memory", self.executor_memory)
            .config("spark.driver.memory", self.driver_memory)
            .getOrCreate()
        )

        self.jspark = self.spark._jsparkSession
        self.java_import = None
        self.prepare()
        self.nebula_spark_ds = NEBULA_SPARK_CONNECTOR_DATASOURCE

        # TBD: ping NebulaGraph Meta and Storage Server, fail and guide user to check config

    def __str__(self):
        return f"SparkEngine: {self.spark}"

    def parse_config(self):
        """parse and validate config"""
        if self.config is None:
            return

        if self.config.shuffle_partitions is not None:
            self.shuffle_partitions = self.config.shuffle_partitions

        if self.config.executor_memory is not None:
            self.executor_memory = self.config.executor_memory

        if self.config.driver_memory is not None:
            self.driver_memory = self.config.driver_memory

        if self.config.encode_vertex_id is not None:
            self.encode_vertex_id = self.config.encode_vertex_id

    def prepare(self):
        self.java_import = self._get_java_import()

    def _get_java_import(self, force=False):
        if self.java_import is not None and not force:
            return self.java_import

        from py4j.java_gateway import java_import

        # scala:
        # import "com.vesoft.nebula.algorithm.config.SparkConfig"
        java_import(self.spark._jvm, "com.vesoft.nebula.algorithm.config.SparkConfig")
        return java_import

    def import_scala_class(self, class_name):
        """
        For example:
        scala:
            import "com.vesoft.nebula.algorithm.lib.PageRankAlgo"
        python:
            java_import(spark._jvm, "com.vesoft.nebula.algorithm.lib.PageRankAlgo")
        """
        self.java_import(self.spark._jvm, class_name)

    def import_algo_config_class(self, class_name):
        self.import_scala_class(f"{NEBULA_ALGO_CONFIG}.{class_name}")

    def import_algo_lib_class(self, class_name):
        self.import_scala_class(f"{NEBULA_ALGO_LIB}.{class_name}")


class NebulaEngine(BaseEngine):
    def __init__(self, config=None):
        self.type = "nebula"
        self.config = config
        self.parse_config()

    def __str__(self):
        return f"NebulaEngine: {self.config}"

    def parse_config(self):
        """parse and validate config"""
        if self.config is None:
            return

    def prepare(self):
        pass



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 The NebulaGraph Authors. All rights reserved.

from __future__ import annotations

# Default Configuration
GRAPHD_HOSTS = "graphd:9669"
METAD_HOSTS = "metad0:9559,metad1:9559,metad2:9559"
USER = "root"
PASSWORD = "nebula"
SPACE = "basketballplayer"


class NebulaGraphConfig:
    def __init__(
        self,
        graphd_hosts: str = GRAPHD_HOSTS,
        metad_hosts: str = METAD_HOSTS,
        user: str = USER,
        password: str = PASSWORD,
        space: str = SPACE,
        **kwargs,
    ):
        self.graphd_hosts = graphd_hosts
        self.metad_hosts = metad_hosts
        self.user = user
        self.password = password
        self.space = space
        self.shuffle_partitions = None
        self.executor_memory = None
        self.driver_memory = None
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getattr__(self, name):
        return None



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 The NebulaGraph Authors. All rights reserved.

from __future__ import annotations

import networkx as nx


class NebulaGraphObject:
    def __init__(self, df: NebulaDataFrameObject):
        self.engine = df.engine
        self.df = df
        # if engine is nebula, self._graph is a networkx graph object
        # if engine is spark, self._graph is a spark graph object
        self._graph = None

    def get_engine(self):
        return self.engine

    @property
    def algo(self):
        from ngdi.nebula_algo import NebulaAlgorithm as NebulaAlgorithmImpl

        return NebulaAlgorithmImpl(self)

    def get_nx_graph(self):
        if self.engine.type == "nebula":
            if self._graph is None:
                # convert the graph to a networkx graph
                # and return the result
                self.to_nx_graph()
            return self._graph
        else:
            # for now the else case will be spark, to networkx is not supported
            raise Exception(
                "For NebulaGraphObject in spark engine,"
                "convert to networkx graph is not supported",
            )

    def to_networkx(self, update=False):
        if self.engine.type == "nebula":
            # convert the graph to a networkx graph
            # and return the result
            if self._graph is None or update:
                self._graph = self.df.to_networkx()
            return self._graph
        else:
            # for now the else case will be spark, to networkx is not supported
            raise Exception(
                "For NebulaGraphObject in spark engine,"
                "convert to networkx graph is not supported",
            )

    def to_graphx(self, update=False):
        if self.engine.type == "spark":
            # convert the graph to a graphx graph
            # and return the result
            if self._graph is None or update:
                self._graph = self.df.to_graphx()
            return self._graph
        else:
            # for now the else case will be nebula, to graphx is not supported
            raise Exception(
                "For NebulaGraphObject in nebula engine,"
                "convert to graphx is not supported",
            )


class NebulaDataFrameObject:
    def __init__(self, engine, data):
        """
        data: pd.DataFrame or spark.DataFrame
        """
        self.engine = engine
        self.data = data
        # if engine is nebula, self.data is a pandas dataframe
        # if engine is spark, self.data is a spark dataframe

    def get_engine(self):
        return self.engine

    @property
    def algo(self):
        from ngdi.nebula_algo import NebulaAlgorithm as NebulaAlgorithmImpl

        return NebulaAlgorithmImpl(self)

    def to_spark_df(self):
        if self.engine.type == "spark":
            return self.data
        else:
            # convert pandas dataframe to spark dataframe
            # not implemented now
            raise NotImplementedError

    def to_pandas_df(self):
        if self.engine.type == "nebula":
            return self.data
        else:
            # convert the spark df to a pandas data frame
            # we should carefully think about this use case
            # not implemented now
            raise NotImplementedError

    def to_networkx(self):
        if self.engine.type == "nebula":
            return nx.from_pandas_edgelist(self.data, "src", "dst")
        else:
            # for now the else case will be spark, to networkx is not supported
            raise Exception(
                "For NebulaDataFrameObject in spark engine,"
                "convert to networkx graph is not supported",
            )

    def to_graphx(self):
        if self.engine.type == "spark":
            df = self.data
            # convert the df to a graphx graph, not implemented now
            raise NotImplementedError
        else:
            # for now the else case will be nebula, to graphx is not supported
            raise Exception(
                "For NebulaDataFrameObject in nebula engine,"
                "convert to graphx is not supported",
            )

    def to_graph(self):
        return NebulaGraphObject(self)

    def show(self, *keywords, **kwargs):
        return self.data.show(*keywords, **kwargs)



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 The NebulaGraph Authors. All rights reserved.
from __future__ import annotations

from ngdi.config import NebulaGraphConfig
from ngdi.nebula_data import NebulaDataFrameObject

SPARK_NEBULA_SINKS = ["nebulagraph_vertex", "nebulagraph_edge"]
SPARK_FILE_SINKS = ["csv", "json", "parquet"]
SPARK_SERVER_SINKS = []
SPARK_SINKS = SPARK_NEBULA_SINKS + SPARK_FILE_SINKS + SPARK_SERVER_SINKS


class NebulaWriterBase(object):
    def __init__(self, engine=None, config=None, **kwargs):
        self.engine_type = engine
        self.config = config

    def write(self, **kwargs):
        raise NotImplementedError

    def options(self, **kwargs):
        raise NotImplementedError

    def show_options(self, **kwargs):
        raise NotImplementedError


class NebulaWriter:
    def __init__(
        self,
        data,
        sink: str = "nebulagraph_vertex",
        config: NebulaGraphConfig = NebulaGraphConfig(),
        engine: str = "spark",
        **kwargs,
    ):
        self.engine_type = engine
        if self.engine_type == "spark":
            self.writer = NebulaWriterWithSpark(data, sink, config, **kwargs)
        elif self.engine_type == "nebula":
            self.writer = NebulaWriterWithGraph(data, sink, config, **kwargs)
        else:
            raise NotImplementedError

    def __getattr__(self, name):
        return getattr(self.writer, name)


class NebulaWriterWithGraph(NebulaWriterBase):
    def __init__(self, data, sink: str, config: NebulaGraphConfig, **kwargs):
        super().__init__("nebula", config, **kwargs)
        from ngdi.engines import NebulaEngine

        self.engine = NebulaEngine(config)
        self.raw_df = None
        self.df = None


class NebulaWriterWithSpark(NebulaWriterBase):
    def __init__(self, data, sink: str, config: NebulaGraphConfig, **kwargs):
        super().__init__("spark", config, **kwargs)
        from ngdi.engines import SparkEngine

        self.engine = SparkEngine(config)
        self.raw_df = None
        self.raw_df_writer = None
        self.df = None
        self.jdf = None
        self.sink = sink
        self._options = {}

        from py4j.java_gateway import JavaObject

        if isinstance(data, NebulaDataFrameObject):
            self.df = data
            self.raw_df = data.data
        elif isinstance(data, JavaObject):
            self.jdf = data
        else:
            raise NotImplementedError

    def _convert_java_object_to_spark_df(self, jdf, column_map: dict = {}):
        """
        Convert JavaObject to Spark DataFrame, with column name mapping
        ref: https://stackoverflow.com/questions/36023860/how-to-use-a-scala-class-inside-pyspark
        """
        from pyspark.sql import DataFrame

        spark = self.engine.spark
        jdf = jdf.toDF()
        for k, v in column_map.items():
            jdf = jdf.withColumnRenamed(k, v)
        return DataFrame(jdf, spark)

    def _get_raw_df_writer_with_nebula(self):
        datasource_format = self.engine.nebula_spark_ds
        # hard code as overwrite SaveMode
        return self.raw_df.write.format(datasource_format).mode("overwrite")

    def _get_raw_df_writer_with_file(self):
        raise NotImplementedError

    def _get_raw_df_writer_with_server(self):
        raise NotImplementedError

    def _get_raw_df_writer(self):
        # validate sink
        assert self.sink in SPARK_SINKS, (
            f"Sink should be one of {SPARK_SINKS}, " f"but got {self.sink}",
        )

        if self.raw_df_writer is None:
            if self.sink in SPARK_NEBULA_SINKS:
                self.raw_df_writer = self._get_raw_df_writer_with_nebula()
            elif self.sink in SPARK_FILE_SINKS:
                self.raw_df_writer = self._get_raw_df_writer_with_file()
            elif self.sink in SPARK_SERVER_SINKS:
                self.raw_df_writer = self._get_raw_df_writer_with_server()
            else:
                raise NotImplementedError
        return self.raw_df_writer

    def _set_options_with_nebula(self, **kwargs):
        writer = self.raw_df_writer
        # type setting
        type = kwargs.get("type", "vertex")
        assert type in [
            "vertex",
            "edge",
        ], f"type should be vertex or edge, but got {type}"
        writer.option("type", type)
        self._options["type"] = type

        # space setting, set only when it's specified, else inherit from NebulaGraphConfig
        space = kwargs.get("space", self.config.space)
        assert space is not None, "space should be specified"
        writer.option("spaceName", space)
        self._options["spaceName"] = space

        # label setting, for vertex case, it's tag, for edge case, it's edge type
        label = (
            kwargs.get("tag", None)
            if type == "vertex"
            else kwargs.get("edge_type", None)
        )
        assert (
            label is not None
        ), "tag(for vertex) or edge_type(for edge) should be specified"
        writer.option("label", label)
        self._options["label"] = label

        # writeMode setting, by default, it's insert, accept update or delete, too
        write_mode = kwargs.get("write_mode", "insert")
        assert write_mode in [
            "insert",
            "update",
            "delete",
        ], f"write_mode should be insert, update or delete, but got {write_mode}"
        writer.option("writeMode", write_mode)

        # batch setting, by default, it's 256
        batch = kwargs.get("batch", 256)
        assert isinstance(
            batch, int
        ), f"batch should be an integer, but got {type(batch)}"
        writer.option("batch", batch)
        self._options["batch"] = batch

        # credential from NebulaGraphConfig
        writer.option("metaAddress", self.config.metad_hosts)
        writer.option("graphAddress", self.config.graphd_hosts)
        writer.option("user", self.config.user)
        self._options["metaAddress"] = self.config.metad_hosts
        self._options["graphAddress"] = self.config.graphd_hosts
        self._options["user"] = self.config.user
        writer.option("password", self.config.password)

        if type == "vertex":
            # vidPolicy setting, by default, it's empty, accept hash or uuid, too
            vid_policy = kwargs.get("vid_policy", "")
            assert vid_policy in [
                "",
                "hash",
                "uuid",
            ], f"vid_policy should be empty, hash or uuid, but got {vid_policy}"
            writer.option("vidPolicy", vid_policy)
            self._options["vidPolicy"] = vid_policy

            # vertexField setting, by default, it's _id, must be a string
            vertexField = kwargs.get("vid_field", "_id")
            assert isinstance(
                vertexField, str
            ), f"vid_field should be a string, but got {type(vertexField)}"
            writer.option("vertexField", vertexField)
            self._options["vertexField"] = vertexField

        if type == "edge":
            # srcId setting, by default, it's srcId, must be a string
            src_id = kwargs.get("src_id", "srcId")
            assert isinstance(
                src_id, str
            ), f"src_id should be a string, but got {type(src_id)}"
            writer.option("srcId", src_id)
            self._options["srcId"] = src_id

            # dstId setting, by default, it's dstId, must be a string
            dst_id = kwargs.get("dst_id", "dstId")
            assert isinstance(
                dst_id, str
            ), f"dst_id should be a string, but got {type(dst_id)}"
            writer.option("dstId", dst_id)
            self._options["dstId"] = dst_id

            # srcIdPolicy setting, by default, it's empty, accept hash or uuid, too
            src_id_policy = kwargs.get("src_id_policy", "")
            assert src_id_policy in [
                "",
                "hash",
                "uuid",
            ], f"src_id_policy should be empty, hash or uuid, but got {src_id_policy}"
            writer.option("srcIdPolicy", src_id_policy)
            self._options["srcIdPolicy"] = src_id_policy

            # dstIdPolicy setting, by default, it's empty, accept hash or uuid, too
            dst_id_policy = kwargs.get("dst_id_policy", "")
            assert dst_id_policy in [
                "",
                "hash",
                "uuid",
            ], f"dst_id_policy should be empty, hash or uuid, but got {dst_id_policy}"
            writer.option("dstIdPolicy", dst_id_policy)
            self._options["dstIdPolicy"] = dst_id_policy

            # randkField setting, by default, it's empty, must be a string
            rank_field = kwargs.get("rank_field", "")
            assert isinstance(
                rank_field, str
            ), f"rank_field should be a string, but got {type(rank_field)}"
            writer.option("rankField", rank_field)
            self._options["rankField"] = rank_field

    def _set_options_with_file(self, **kwargs):
        raise NotImplementedError

    def _set_options_with_server(self, **kwargs):
        raise NotImplementedError

    def set_options(self, **kwargs):
        """
        example:
        # change col name
        louvain_result_df = louvain_result.toDF()

        # CREATE TAG IF NOT EXISTS louvain (
        #     cluster_id string NOT NULL
        # );

        louvain_result_df = louvain_result_df.withColumnRenamed("louvain", "cluster_id")
        louvain_result_df.write.format("com.vesoft.nebula.connector.NebulaDataSource").option(
            "type", "vertex").option(
            "spaceName", "basketballplayer").option(
            "label", "louvain").option(
            "vidPolicy", "").option(
            "vertexField", "_id").option(
            "batch", 1).option(
            "metaAddress", "metad0:9559,metad1:9559,metad2:9559").option(
            "graphAddress", "graphd:9669").option(
            "passwd", "nebula").option(
            "user", "root").option(
            "writeType", "insert")
        """
        if self.raw_df is None:
            if self.jdf is None:
                raise ValueError("DataFrame is None in NebulaWriter")
            self.raw_df = self._convert_java_object_to_spark_df(
                self.jdf, kwargs.get("properties", {})
            )
        # TBD: check tag/edge type's schema against df's schema, raise error if not match

        self._get_raw_df_writer()  # self.raw_df_writer

        # case switch based on sink
        if self.sink in SPARK_NEBULA_SINKS:
            self._set_options_with_nebula(**kwargs)
        elif self.sink in SPARK_FILE_SINKS:
            self._set_options_with_file(**kwargs)
        elif self.sink in SPARK_SERVER_SINKS:
            self._set_options_with_server(**kwargs)
        else:
            raise NotImplementedError

        return self._options

    def get_options(self):
        return self._options

    def write(self):
        # Check self.raw_df, if it is None, raise exception
        if self.raw_df_writer is None:
            raise Exception(
                "No options set for NebulaWriter, please call options() first"
            )
        return self.raw_df_writer.save()



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 The NebulaGraph Authors. All rights reserved.

from __future__ import annotations

from ngdi.nebula_data import NebulaGraphObject as NebulaGraphObjectImpl
from ngdi.nebula_data import NebulaDataFrameObject as NebulaDataFrameObjectImpl


def algo(func):
    func.is_algo = True
    return func


class NebulaAlgorithm:
    def __init__(self, obj: NebulaGraphObjectImpl or NebulaDataFrameObjectImpl):
        if isinstance(obj, NebulaGraphObjectImpl):
            self.algorithm = NebulaGraphAlgorithm(obj)
        elif isinstance(obj, NebulaDataFrameObjectImpl):
            self.algorithm = NebulaDataFrameAlgorithm(obj)
        else:
            raise ValueError(
                f"Unsupported object type: {type(obj)}",
                "NebulaAlgorithm only supports NebulaGraphObject and"
                "NebulaDataFrameObject",
            )
        self.obj = obj
        self.engine = obj.engine

    def __getattr__(self, name):
        return getattr(self.algorithm, name)


class NebulaDataFrameAlgorithm:
    """
    Spark Dataframe to run algorithm
    """

    def __init__(self, ndf_obj: NebulaDataFrameObjectImpl):
        self.ndf_obj = ndf_obj
        self.algorithms = []

    def register_algo(self, func):
        self.algorithms.append(func.__name__)

    def get_all_algo(self):
        if not self.algorithms:
            for name, func in NebulaDataFrameAlgorithm.__dict__.items():
                if hasattr(func, "is_algo"):
                    self.register_algo(func)
        return self.algorithms

    def check_engine(self):
        """
        Check if the engine is supported.
        For netowrkx, we need to convert the NebulaDataFrameObject to NebulaGraphObject
        For spark, we can directly use the NebulaDataFrameObject
        """
        if self.ndf_obj.engine.type == "networkx":
            raise Exception(
                "For NebulaDataFrameObject in networkx engine,"
                "Please transform it to NebulaGraphObject to run algorithm",
                "For example: g = nebula_df.to_graph; g.algo.pagerank()",
            )

    def get_spark_engine_context(self, config_class: str, lib_class: str):
        """
        Get the engine context
        """
        self.check_engine()
        engine = self.ndf_obj.engine
        spark = engine.spark
        jspark = engine.jspark
        engine.import_algo_config_class(config_class)
        engine.import_algo_lib_class(lib_class)
        return engine, spark, jspark, engine.encode_vertex_id

    def get_spark_dataframe(self):
        """
        Check if df is a pyspark.sql.dataframe.DataFrameme, and return it
        """
        df = self.ndf_obj.data
        from pyspark.sql.dataframe import DataFrame as pyspark_sql_df

        if not isinstance(df, pyspark_sql_df):
            raise Exception(
                "The NebulaDataFrameObject is not a spark dataframe",
                f"Got type(df): {type(df)}",
            )
        return df

    @algo
    def pagerank(
        self, reset_prob: float = 0.15, max_iter: int = 10, weighted: bool = False
    ):
        engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
            "PRConfig", "PageRankAlgo"
        )
        df = self.get_spark_dataframe()
        config = spark._jvm.PRConfig(max_iter, reset_prob, encode_vertex_id)
        result = spark._jvm.PageRankAlgo.apply(jspark, df._jdf, config, weighted)

        return result

    @algo
    def connected_components(self, max_iter: int = 10, weighted: bool = False):
        engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
            "CcConfig", "ConnectedComponentsAlgo"
        )
        df = self.get_spark_dataframe()
        config = spark._jvm.CcConfig(max_iter, encode_vertex_id)
        result = spark._jvm.ConnectedComponentsAlgo.apply(
            jspark, df._jdf, config, weighted
        )

        return result

    @algo
    def label_propagation(self, max_iter: int = 10, weighted: bool = False):
        engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
            "LPAConfig", "LabelPropagationAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.LPAConfig(max_iter, encode_vertex_id)
        result = spark._jvm.LabelPropagationAlgo.apply(
            jspark, df._jdf, config, weighted
        )

        return result

    @algo
    def louvain(self, max_iter: int = 20, internalIter: int = 10, tol: float = 0.5):
        engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
            "LouvainConfig", "LouvainAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.LouvainConfig(max_iter, internalIter, tol, encode_vertex_id)
        result = spark._jvm.LouvainAlgo.apply(jspark, df._jdf, config, False)

        return result

    @algo
    def k_core(self, max_iter: int = 10, degree: int = 2):
        engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
            "KCoreConfig", "KCoreAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.KCoreConfig(max_iter, degree, encode_vertex_id)

        result = spark._jvm.KCoreAlgo.apply(jspark, df._jdf, config)

        return result

    # def shortest_path(self, landmarks: list, weighted: bool = False):
    #     engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
    #         "ShortestPathConfig", "ShortestPathAlgo"
    #     )
    #     # TBD: ShortestPathAlgo is not yet encodeID compatible
    #     df = self.get_spark_dataframe()

    #     config = spark._jvm.ShortestPathConfig(landmarks, encode_vertex_id)
    #     result = spark._jvm.ShortestPathAlgo.apply(jspark, df._jdf, config, weighted)

    #     return result

    @algo
    def degree_statics(self):
        engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
            "DegreeStaticConfig", "DegreeStaticAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.DegreeStaticConfig(encode_vertex_id)
        result = spark._jvm.DegreeStaticAlgo.apply(jspark, df._jdf, config)

        return result

    @algo
    def betweenness_centrality(
        self, max_iter: int = 10, degree: int = 2, weighted: bool = False
    ):
        engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
            "BetweennessConfig", "BetweennessCentralityAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.BetweennessConfig(max_iter, encode_vertex_id)
        result = spark._jvm.BetweennessCentralityAlgo.apply(
            jspark, df._jdf, config, weighted
        )

        return result

    @algo
    def coefficient_centrality(self, type: str = "local"):
        # type could be either "local" or "global"
        assert type.lower() in ["local", "global"], (
            "type should be either local or global"
            f"in coefficient_centrality algo. Got type: {type}"
        )
        engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
            "CoefficientConfig", "ClusteringCoefficientAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.CoefficientConfig(type, encode_vertex_id)
        result = spark._jvm.ClusteringCoefficientAlgo.apply(jspark, df._jdf, config)

        return result

    @algo
    def bfs(self, max_depth: int = 10, root: int = 1):
        engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
            "BfsConfig", "BfsAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.BfsConfig(max_depth, root, encode_vertex_id)
        result = spark._jvm.BfsAlgo.apply(jspark, df._jdf, config)

        return result

    # dfs is not yet supported, need to revisit upstream nebula-algorithm
    @algo
    def dfs(self, max_depth: int = 10, root: int = 1):
        engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
            "DfsConfig", "DfsAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.DfsConfig(max_depth, root, encode_vertex_id)
        result = spark._jvm.DfsAlgo.apply(jspark, df._jdf, config)

        return result

    @algo
    def hanp(
        self,
        hop_attenuation: float = 0.5,
        max_iter: int = 10,
        preference: float = 1.0,
        weighted: bool = False,
        preferences=None,
    ):
        engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
            "HanpConfig", "HanpAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.HanpConfig(
            hop_attenuation, max_iter, preference, encode_vertex_id
        )
        result = spark._jvm.HanpAlgo.apply(
            jspark, df._jdf, config, weighted, preferences
        )

        return result

    # @algo
    # def node2vec(
    #     self,
    #     max_iter: int = 10,
    #     lr: float = 0.025,
    #     data_num_partition: int = 10,
    #     model_num_partition: int = 10,
    #     dim: int = 10,
    #     window: int = 3,
    #     walk_length: int = 5,
    #     num_walks: int = 3,
    #     p: float = 1.0,
    #     q: float = 1.0,
    #     directed: bool = False,
    #     degree: int = 30,
    #     emb_separator: str = ",",
    #     model_path: str = "hdfs://127.0.0.1:9000/model",
    #     weighted: bool = False,
    # ):
    #     engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
    #         "Node2vecConfig", "Node2VecAlgo"
    #     )
    #     # TBD: Node2VecAlgo is not yet encodeID compatible
    #     df = self.get_spark_dataframe()
    #     config = spark._jvm.Node2vecConfig(
    #         max_iter,
    #         lr,
    #         data_num_partition,
    #         model_num_partition,
    #         dim,
    #         window,
    #         walk_length,
    #         num_walks,
    #         p,
    #         q,
    #         directed,
    #         degree,
    #         emb_separator,
    #         model_path,
    #         encode_vertex_id,
    #     )
    #     result = spark._jvm.Node2VecAlgo.apply(jspark, df._jdf, config, weighted)

    #     return result

    @algo
    def jaccard(self, tol: float = 1.0):
        engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
            "JaccardConfig", "JaccardAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.JaccardConfig(tol, encode_vertex_id)
        result = spark._jvm.JaccardAlgo.apply(jspark, df._jdf, config)

        return result

    @algo
    def strong_connected_components(self, max_iter: int = 10, weighted: bool = False):
        engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
            "CcConfig", "StronglyConnectedComponentsAlgo"
        )
        df = self.get_spark_dataframe()
        config = spark._jvm.CcConfig(max_iter, encode_vertex_id)
        result = spark._jvm.StronglyConnectedComponentsAlgo.apply(
            jspark, df._jdf, config, weighted
        )

        return result

    @algo
    def triangle_count(self):
        engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
            "TriangleConfig", "TriangleCountAlgo"
        )
        df = self.get_spark_dataframe()
        config = spark._jvm.TriangleConfig(encode_vertex_id)
        result = spark._jvm.TriangleCountAlgo.apply(jspark, df._jdf, config)

        return result

    # @algo
    # def closeness(self, weighted: bool = False):
    #     # TBD: ClosenessAlgo is not yet encodeID compatible
    #     engine, spark, jspark, encode_vertex_id = self.get_spark_engine_context(
    #         "ClosenessConfig", "ClosenessAlgo"
    #     )
    #     df = self.get_spark_dataframe()
    #     config = spark._jvm.ClosenessConfig(weighted, encode_vertex_id)
    #     result = spark._jvm.ClosenessAlgo.apply(jspark, df._jdf, config, False)

    #     return result


class NebulaGraphAlgorithm:
    """
    Networkx to run algorithm
    """

    def __init__(self, graph):
        self.graph = graph
        self.algorithms = []

    def register_algo(self, func):
        self.algorithms.append(func.__name__)

    def get_all_algo(self):
        if not self.algorithms:
            for name, func in NebulaGraphAlgorithm.__dict__.items():
                if hasattr(func, "is_algo"):
                    self.register_algo(func)
        return self.algorithms

    def check_engine(self):
        """
        Check if the engine is supported.
        For netowrkx, we can directly call .algo.pagerank()
        For spark, we need to convert the NebulaGraphObject to NebulaDataFrameObject
        """
        if self.graph.engine.type == "spark":
            raise Exception(
                "For NebulaGraphObject in spark engine,"
                "Please transform it to NebulaDataFrameObject to run algorithm",
                "For example: df = nebula_graph.to_df; df.algo.pagerank()",
            )

    @algo
    def pagerank(self, reset_prob=0.15, max_iter=10):
        self.check_engine()
        pass



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 The NebulaGraph Authors. All rights reserved.


class NebulaGNN:
    def __init__(self, engine=None, config=None, **kwargs):
        self.engine = engine
        self.config = config



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 The NebulaGraph Authors. All rights reserved.

from pkgutil import extend_path


__path__ = extend_path(__path__, __name__)  # type: ignore

from ngdi.nebula_reader import NebulaReader
from ngdi.nebula_writer import NebulaWriter
from ngdi.nebula_algo import NebulaAlgorithm
from ngdi.nebula_gnn import NebulaGNN
from ngdi.config import NebulaGraphConfig

# export
__all__ = (
    "NebulaReader",
    "NebulaWriter",
    "NebulaAlgorithm",
    "NebulaGNN",
    "NebulaGraphConfig",
)



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 The NebulaGraph Authors. All rights reserved.
from __future__ import annotations

from ngdi.config import NebulaGraphConfig
from ngdi.nebula_data import NebulaDataFrameObject


class NebulaReaderBase(object):
    def __init__(self, engine=None, config=None, **kwargs):
        self.engine_type = engine
        self.config = config

    def scan(self, **kwargs):
        raise NotImplementedError

    def query(self, **kwargs):
        raise NotImplementedError

    def load(self, **kwargs):
        raise NotImplementedError

    def read(self, **kwargs):
        raise NotImplementedError

    def show(self, **kwargs):
        raise NotImplementedError


class NebulaReader:
    def __init__(
        self, engine="spark", config: NebulaGraphConfig = NebulaGraphConfig(), **kwargs
    ):
        self.engine_type = engine
        if self.engine_type == "spark":
            self.reader = NebulaReaderWithSpark(config, **kwargs)
        elif self.engine_type == "nebula":
            self.reader = NebulaReaderWithGraph(config, **kwargs)
        else:
            raise NotImplementedError

    def __getattr__(self, name):
        return getattr(self.reader, name)


class NebulaReaderWithGraph(NebulaReaderBase):
    def __init__(self, config: NebulaGraphConfig, **kwargs):
        super().__init__("nebula", config, **kwargs)
        from ngdi.engines import NebulaEngine

        self.engine = NebulaEngine(config)
        self.raw_df = None
        self.df = None

    def scan(self, **kwargs):
        # Implement the scan method specific to Nebula engine
        raise NotImplementedError

    def query(self, **kwargs):
        # Implement the query method specific to Nebula engine
        raise NotImplementedError

    def load(self, **kwargs):
        # Implement the load method specific to Nebula engine
        raise NotImplementedError

    def read(self, **kwargs):
        # Implement the read method specific to Nebula engine
        raise NotImplementedError

    def show(self, **kwargs):
        # Implement the show method specific to Nebula engine
        raise NotImplementedError


class NebulaReaderWithSpark(NebulaReaderBase):
    def __init__(self, config: NebulaGraphConfig, **kwargs):
        super().__init__("spark", config, **kwargs)
        from ngdi.engines import SparkEngine

        self.engine = SparkEngine(config)
        self.raw_df = None
        self.raw_df_reader = None
        self.df = None

    def scan(self, **kwargs):
        """
        example:
        df = spark.read.format(
        "com.vesoft.nebula.connector.NebulaDataSource").option(
            "type", "edge").option(
            "spaceName", "basketballplayer").option(
            "label", "follow").option(
            "returnCols", "degree").option(
            "metaAddress", "metad0:9559").option(
            "partitionNumber", 3).load()
        """

        # validate kwargs, there should be:
        # - edge: edge type, string
        # - props: properties to be returned, string

        assert "edge" in kwargs, "edge type should be specified"
        assert "props" in kwargs, "properties to be returned should be specified"

        # validate config, there should be:
        # - metad_hosts: meta server address, string
        # - space: space name, string
        assert self.config.metad_hosts, "metad_hosts should be specified"
        assert self.config.space, "space should be specified"
        edge_type = kwargs["edge"]
        props = kwargs["props"]
        partition_number = kwargs.get("partition_number", 3)  # default 3

        space_name = self.config.space
        metad_hosts = self.config.metad_hosts

        spark = self.engine.spark
        datasource_format = self.engine.nebula_spark_ds

        self.raw_df_reader = (
            spark.read.format(datasource_format)
            .option("type", "edge")
            .option("spaceName", space_name)
            .option("label", edge_type)
            .option("returnCols", props)
            .option("metaAddress", metad_hosts)
            .option("partitionNumber", partition_number)
        )

    def query(self, query: str, **kwargs):
        # Implement the query method specific to Spark engine
        """
        df = spark.read.format(
        "com.vesoft.nebula.connector.NebulaDataSource").option(
            "type", "edge").option(
            "spaceName", "basketballplayer").option(
            "label", "follow").option(
            "returnCols", "degree").option(
            "metaAddress", "metad0:9559").option(
            "graphAddress", "graphd:9669").option(
            "ngql", "MATCH ()-[e:follow]->() return e LIMIT 1000").option(
            "partitionNumber", 1).load()
        """

        # validate kwargs, there should be:
        # - edge: edge type, string
        # - props: properties to be returned, string
        # - ngql: ngql to be executed, string

        assert "edge" in kwargs, "edge type should be specified"
        assert "props" in kwargs, "properties to be returned should be specified"
        assert query is not None, "ngql should be specified"

        # validate config, there should be:
        # - metad_hosts: meta server address, string
        # - space: space name, string
        # - graphd_hosts: graph server address, string
        assert self.config.metad_hosts, "metad_hosts should be specified"
        assert self.config.space, "space should be specified"
        assert self.config.graphd_hosts, "graphd_hosts should be specified"
        edge_type = kwargs["edge"]
        props = kwargs["props"]

        space_name = self.config.space
        metad_hosts = self.config.metad_hosts
        graphd_hosts = self.config.graphd_hosts

        spark = self.engine.spark
        datasource_format = self.engine.nebula_spark_ds

        self.raw_df_reader = (
            spark.read.format(datasource_format)
            .option("type", "edge")
            .option("spaceName", space_name)
            .option("label", edge_type)
            .option("returnCols", props)
            .option("metaAddress", metad_hosts)
            .option("graphAddress", graphd_hosts)
            .option("ngql", query)
            .option("partitionNumber", 1)
        )

    def load(self, **kwargs):
        # Implement the load method specific to Spark engine
        raise NotImplementedError

    def read(self, **kwargs):
        # Check self.raw_df, if it is None, raise exception
        if self.raw_df_reader is None:
            raise Exception("No data loaded, please use scan or query first")
        self.raw_df = self.raw_df_reader.load()
        self.df = NebulaDataFrameObject(engine=self.engine, data=self.raw_df)
        return self.df

    def show(self, **kwargs):
        # Implement the show method specific to Spark engine
        raise NotImplementedError
